{% block container %}
{% endblock %}

{% block start_script %}
    {% if chart.tag_script_js %}
    <script>
    {% endif %}

    {% if chart.jquery_on_ready %}
    $(function(){
    {% endif %}
{% endblock start_script %}

{% block body %}

    {% block data %}
    {% endblock data %}

    {% block inject %}
    {% endblock inject %}

    {% block foot %}
    {% endblock foot %}
     <!-- end of body from base-->

     {% endblock body%}

{% block endscript %}
    {% if chart.jquery_on_ready %}
        });
    {% endif %}
    {% if chart.tag_script_js %}
    </script>
    {% endif %}
{% endblock endscript %}
