{# This template adds attributes unique
   to lineWithFocusChart #}

{% extends "content.html" %}
{% block body %}

{# calling super guarantees everying in content is also found here ...#}
{{super()}}

{% endblock body %}
