{#
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
#}
{% extends "airflow/task_instance.html" %}
{% block title %}Airflow - DAGs{% endblock %}

{% block content %}
    {{ super() }}
    <h4>{{ title }}</h4>
    <div>
        <table class="table table-striped table-bordered">
            <tr>
                <th>Key</th>
                <th>Value</th>
            </tr>
            {% for attr, value in attributes %}
                <tr>
                    <td>{{ attr }}</td>
                    <td class='code'>{{ value }}</td>
                </tr>
            {% endfor %}
        </table>
        {{ html_code|safe }}
    </div>
{% endblock %}
