{# This template adds attributes unique
   to linePlusBarChart #}

{% extends "content.html" %}
{% block body %}

    {% block data %}
    {{super()}}
    {% endblock data %}

    {% block init %}
    {{super()}}
    {% endblock init %}

    {% block focus %}
    {% if chart.focus_enable %}
        chart.focusEnable(true);
    {% else %}
        chart.focusEnable(false);
    {% endif %}
    {% endblock focus %}

    {% block axes %}
    {{super()}}
    {% endblock axes %}

    {% block legend %}
    {{super()}}
    {% endblock legend %}

    {% block custoattr %}
    {{super()}}
    {% endblock custoattr %}

    {% block inject %}
    {{ super() }}
    {% endblock inject %}

    {% block close %}
    {{ super() }}
    {% endblock close %}

{% endblock body %}
